==> openpdf-1.2.17-sources.jar: com/lowagie/text/pdf/PdfNameTree.java <==
/*
 * Copyright 2004 by Paulo Soares.
 *
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the License.
 *
 * The Original Code is 'iText, a free JAVA-PDF library'.
 *
 * The Initial Developer of the Original Code is Bruno Lowagie. Portions created by
 * the Initial Developer are Copyright (C) 1999, 2000, 2001, 2002 by Bruno Lowagie.
 * All Rights Reserved.
 * Co-Developer of the code is Paulo Soares. Portions created by the Co-Developer
 * are Copyright (C) 2000, 2001, 2002 by Paulo Soares. All Rights Reserved.
 *
 * Contributor(s): all the names of the contributors are added in the source code
 * where applicable.
 *
 * Alternatively, the contents of this file may be used under the terms of the
 * LGPL license (the "GNU LIBRARY GENERAL PUBLIC LICENSE"), in which case the
 * provisions of LGPL are applicable instead of those above.  If you wish to
 * allow use of your version of this file only under the terms of the LGPL
 * License and not to allow others to use your version of this file under
 * the MPL, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the LGPL.
 * If you do not delete the provisions above, a recipient may use your version
 * of this file under either the MPL or the GNU LIBRARY GENERAL PUBLIC LICENSE.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the MPL as stated above or under the terms of the GNU
 * Library General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Library general Public License for more
 * details.
 *
 * If you didn't download this code from the following link, you should check if
 * you aren't using an obsolete version:
 * http://www.lowagie.com/iText/
 */
package com.lowagie.text.pdf;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;

/**
 * Creates a name tree.
 *
 * @author Paulo Soares (psoares@consiste.pt)
 */
public class PdfNameTree {

  private static final int leafSize = 64;

  /**
   * Writes a name tree to a PdfWriter.
   *
   * @param items the item of the name tree. The key is a <CODE>String</CODE> and the value is a <CODE>PdfObject</CODE>. Note that although
   * the keys are strings only the lower byte is used and no check is made for chars with the same lower byte and different upper byte. This
   * will generate a wrong tree name.
   * @param writer the writer
   * @return the dictionary with the name tree. This dictionary is the one generally pointed to by the key /Dests, for example
   * @throws IOException on error
   */
  public static PdfDictionary writeTree(HashMap items, PdfWriter writer) throws IOException {
    if (items.isEmpty()) {
      return null;
    }
    String[] names = new String[items.size()];
    names = (String[]) items.keySet().toArray(names);
    Arrays.sort(names);
    if (names.length <= leafSize) {
      PdfDictionary dic = new PdfDictionary();
      PdfArray ar = new PdfArray();
      for (String name : names) {
        ar.add(new PdfString(name, null));
        ar.add((PdfObject) items.get(name));
      }
      dic.put(PdfName.NAMES, ar);
      return dic;
    }
    int skip = leafSize;
    PdfIndirectReference[] kids = new PdfIndirectReference[(names.length + leafSize - 1) / leafSize];
    for (int k = 0; k < kids.length; ++k) {
      int offset = k * leafSize;
      int end = Math.min(offset + leafSize, names.length);
      PdfDictionary dic = new PdfDictionary();
      PdfArray arr = new PdfArray();
      arr.add(new PdfString(names[offset], null));
      arr.add(new PdfString(names[end - 1], null));
      dic.put(PdfName.LIMITS, arr);
      arr = new PdfArray();
      for (; offset < end; ++offset) {
        arr.add(new PdfString(names[offset], null));
        arr.add((PdfObject) items.get(names[offset]));
      }
      dic.put(PdfName.NAMES, arr);
      kids[k] = writer.addToBody(dic).getIndirectReference();
    }
    int top = kids.length;
    while (true) {
      if (top <= leafSize) {
        PdfArray arr = new PdfArray();
        for (int k = 0; k < top; ++k) {
          arr.add(kids[k]);
        }
        PdfDictionary dic = new PdfDictionary();
        dic.put(PdfName.KIDS, arr);
        return dic;
      }
      skip *= leafSize;
      int tt = (names.length + skip - 1) / skip;
      for (int k = 0; k < tt; ++k) {
        int offset = k * leafSize;
        int end = Math.min(offset + leafSize, top);
        PdfDictionary dic = new PdfDictionary();
        PdfArray arr = new PdfArray();
        arr.add(new PdfString(names[k * skip], null));
        arr.add(new PdfString(names[Math.min((k + 1) * skip, names.length) - 1], null));
        dic.put(PdfName.LIMITS, arr);
        arr = new PdfArray();
        for (; offset < end; ++offset) {
          arr.add(kids[offset]);
        }
        dic.put(PdfName.KIDS, arr);
        kids[k] = writer.addToBody(dic).getIndirectReference();
      }
      top = tt;
    }
  }

  private static void iterateItems(PdfDictionary dic, HashMap<String, PdfObject> items) {
    PdfArray nn = (PdfArray) PdfReader.getPdfObjectRelease(dic.get(PdfName.NAMES));
    if (nn != null) {
      for (int k = 0; k < nn.size(); ++k) {
        PdfString s = (PdfString) PdfReader.getPdfObjectRelease(nn.getPdfObject(k++));
        items.put(PdfEncodings.convertToString(s.getBytes(), null), nn.getPdfObject(k));
      }
    } else if ((nn = (PdfArray) PdfReader.getPdfObjectRelease(dic.get(PdfName.KIDS))) != null) {
      for (int k = 0; k < nn.size(); ++k) {
        PdfDictionary kid = (PdfDictionary) PdfReader.getPdfObjectRelease(nn.getPdfObject(k));
        iterateItems(kid, items);
      }
    }
  }

  public static HashMap<String, PdfObject> readTree(PdfDictionary dic) {
    HashMap<String, PdfObject> items = new HashMap<>();
    if (dic != null) {
      iterateItems(dic, items);
    }
    return items;
  }
}

==> openpdf-1.3.11-sources.jar: com/lowagie/text/pdf/PdfNameTree.java <==
/*
 * Copyright 2004 by Paulo Soares.
 *
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the License.
 *
 * The Original Code is 'iText, a free JAVA-PDF library'.
 *
 * The Initial Developer of the Original Code is Bruno Lowagie. Portions created by
 * the Initial Developer are Copyright (C) 1999, 2000, 2001, 2002 by Bruno Lowagie.
 * All Rights Reserved.
 * Co-Developer of the code is Paulo Soares. Portions created by the Co-Developer
 * are Copyright (C) 2000, 2001, 2002 by Paulo Soares. All Rights Reserved.
 *
 * Contributor(s): all the names of the contributors are added in the source code
 * where applicable.
 *
 * Alternatively, the contents of this file may be used under the terms of the
 * LGPL license (the "GNU LIBRARY GENERAL PUBLIC LICENSE"), in which case the
 * provisions of LGPL are applicable instead of those above.  If you wish to
 * allow use of your version of this file only under the terms of the LGPL
 * License and not to allow others to use your version of this file under
 * the MPL, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the LGPL.
 * If you do not delete the provisions above, a recipient may use your version
 * of this file under either the MPL or the GNU LIBRARY GENERAL PUBLIC LICENSE.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the MPL as stated above or under the terms of the GNU
 * Library General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Library general Public License for more
 * details.
 *
 * If you didn't download this code from the following link, you should check if
 * you aren't using an obsolete version:
 * http://www.lowagie.com/iText/
 */
package com.lowagie.text.pdf;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
 * Creates a name tree.
 *
 * @author Paulo Soares (psoares@consiste.pt)
 */
public class PdfNameTree {

  private static final int leafSize = 64;

  
  /**
   * Writes a name tree to a PdfWriter.
   *
   * @param items the item of the name tree. The key is a <CODE>String</CODE> and the value is a <CODE>PdfObject</CODE>. Note that although
   * the keys are strings only the lower byte is used and no check is made for chars with the same lower byte and different upper byte. This
   * will generate a wrong tree name.
   * @param writer the writer
   * @return the dictionary with the name tree. This dictionary is the one generally pointed to by the key /Dests, for example
   * @throws IOException on error
   */
  public static PdfDictionary writeTree(HashMap<String, ? extends PdfObject> items, PdfWriter writer) throws IOException {
    return writeTree((Map<String, ? extends PdfObject>) items, writer);
  }
  /**
   * Writes a name tree to a PdfWriter.
   *
   * @param items the item of the name tree. The key is a <CODE>String</CODE> and the value is a <CODE>PdfObject</CODE>. Note that although
   * the keys are strings only the lower byte is used and no check is made for chars with the same lower byte and different upper byte. This
   * will generate a wrong tree name.
   * @param writer the writer
   * @return the dictionary with the name tree. This dictionary is the one generally pointed to by the key /Dests, for example
   * @throws IOException on error
   */
  public static PdfDictionary writeTree(Map<String, ? extends PdfObject> items, PdfWriter writer) throws IOException {
    if (items.isEmpty()) {
      return null;
    }
    String[] names = items.keySet().toArray(new String[0]);
    Arrays.sort(names);
    if (names.length <= leafSize) {
      PdfDictionary dic = new PdfDictionary();
      PdfArray ar = new PdfArray();
      for (String name : names) {
        ar.add(new PdfString(name, null));
        ar.add(items.get(name));
      }
      dic.put(PdfName.NAMES, ar);
      return dic;
    }
    int skip = leafSize;
    PdfIndirectReference[] kids = new PdfIndirectReference[(names.length + leafSize - 1) / leafSize];
    for (int k = 0; k < kids.length; ++k) {
      int offset = k * leafSize;
      int end = Math.min(offset + leafSize, names.length);
      PdfDictionary dic = new PdfDictionary();
      PdfArray arr = new PdfArray();
      arr.add(new PdfString(names[offset], null));
      arr.add(new PdfString(names[end - 1], null));
      dic.put(PdfName.LIMITS, arr);
      arr = new PdfArray();
      for (; offset < end; ++offset) {
        arr.add(new PdfString(names[offset], null));
        arr.add(items.get(names[offset]));
      }
      dic.put(PdfName.NAMES, arr);
      kids[k] = writer.addToBody(dic).getIndirectReference();
    }
    int top = kids.length;
    while (true) {
      if (top <= leafSize) {
        PdfArray arr = new PdfArray();
        for (int k = 0; k < top; ++k) {
          arr.add(kids[k]);
        }
        PdfDictionary dic = new PdfDictionary();
        dic.put(PdfName.KIDS, arr);
        return dic;
      }
      skip *= leafSize;
      int tt = (names.length + skip - 1) / skip;
      for (int k = 0; k < tt; ++k) {
        int offset = k * leafSize;
        int end = Math.min(offset + leafSize, top);
        PdfDictionary dic = new PdfDictionary();
        PdfArray arr = new PdfArray();
        arr.add(new PdfString(names[k * skip], null));
        arr.add(new PdfString(names[Math.min((k + 1) * skip, names.length) - 1], null));
        dic.put(PdfName.LIMITS, arr);
        arr = new PdfArray();
        for (; offset < end; ++offset) {
          arr.add(kids[offset]);
        }
        dic.put(PdfName.KIDS, arr);
        kids[k] = writer.addToBody(dic).getIndirectReference();
      }
      top = tt;
    }
  }

  private static void iterateItems(PdfDictionary dic, HashMap<String, PdfObject> items) {
    PdfArray nn = (PdfArray) PdfReader.getPdfObjectRelease(dic.get(PdfName.NAMES));
    if (nn != null) {
      for (int k = 0; k < nn.size(); ++k) {
        PdfString s = (PdfString) PdfReader.getPdfObjectRelease(nn.getPdfObject(k++));
        items.put(PdfEncodings.convertToString(s.getBytes(), null), nn.getPdfObject(k));
      }
    } else if ((nn = (PdfArray) PdfReader.getPdfObjectRelease(dic.get(PdfName.KIDS))) != null) {
      for (int k = 0; k < nn.size(); ++k) {
        PdfDictionary kid = (PdfDictionary) PdfReader.getPdfObjectRelease(nn.getPdfObject(k));
        iterateItems(kid, items);
      }
    }
  }

  public static HashMap<String, PdfObject> readTree(PdfDictionary dic) {
    HashMap<String, PdfObject> items = new HashMap<>();
    if (dic != null) {
      iterateItems(dic, items);
    }
    return items;
  }
}
